//proto_roar.c:

/*
 *      Copyright (C) Philipp 'ph3-der-loewe' Schafft - 2012-2013
 *
 *  This file is part of SavannahIce,
 *  a protocol converter between RoarAudio and Icecast.
 *  See README for details.
 *
 *  This file is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 3
 *  as published by the Free Software Foundation.
 *
 *  RoarAudio is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include "savannahice.h"

static int __handle_new_stream(int client, void ** userdata, struct roar_message * mes) {
 savannahice_mixer_t mixer;
 struct roar_stream s;

 if ( *userdata != NULL )
  return -1;

 if ( roar_stream_m2s(&s, mes) == -1 )
  return -1;

 if ( s.dir != ROAR_DIR_PLAY )
  return -1;

 *userdata = savannahice_stream_new(client, s.dir, s.info.codec);
 if ( *userdata == NULL )
  return -1;

 mixer = savannahice_mixerstore_get_mixer_by_id(g_mixerstore, mes->stream);
 if ( mixer == NULL ) {
  savannahice_stream_unref(*userdata);
  *userdata = NULL;
  return -1;
 }

 if ( savannahice_mixer_attach_stream(mixer, *userdata) == -1 ) {
  savannahice_mixer_unref(mixer);
  savannahice_stream_unref(*userdata);
  *userdata = NULL;
  return -1;
 }

 savannahice_mixer_unref(mixer);

 return 0;
}

static int __handle_exec_stream(struct roar_vio_calls * vio, void ** userdata, struct roar_message * mes) {
 if ( *userdata == NULL )
  return -1;

 if ( (int)mes->stream != savannahice_stream_get_id(*userdata) )
  return -1;

 if ( roar_vio_nonblock(vio, ROAR_SOCKET_NONBLOCK) == -1 )
  return -1;

 savannahice_stream_set_status(*userdata, ROAR_STREAMSTATE_OLD);

 return 0;
}

static int __handle_execed(struct roar_vio_calls * vio, void ** userdata) {
 savannahice_stream_t stream = *userdata;
 struct roar_buffer * buf;
 ssize_t sizereq;
 void * data;

 ROAR_DBG("__handle_execed(vio=%p, userdata=%p{%p}) = ?", vio, userdata, *userdata);

 if ( savannahice_stream_get_alone(stream) )
  return -1;

 ROAR_DBG("__handle_execed(vio=%p, userdata=%p{%p}) = ?", vio, userdata, *userdata);

 sizereq = savannahice_stream_get_bufferspace(stream);
 if ( sizereq == -1 )
  return -1;

 if ( sizereq == 0 ) {
  // support flow control here.
  return 0;
 }

 ROAR_DBG("__handle_execed(vio=%p, userdata=%p{%p}) = ?", vio, userdata, *userdata);

 if ( roar_buffer_new_data(&buf, sizereq, &data) == -1 )
  return -1;

 sizereq = roar_vio_read(vio, data, sizereq);

 ROAR_DBG("__handle_execed(vio=%p, userdata=%p{%p}) = ?", vio, userdata, *userdata);

 if ( sizereq < 1 ) {
  roar_buffer_free(buf);
  return -1;
 } else {
  if ( roar_buffer_set_len(buf, sizereq) == -1 ) {
   roar_buffer_free(buf);
   return -1;
  }
  savannahice_stream_add_buffer(stream, buf);
  roar_buffer_unref(buf);
 }

 ROAR_DBG("__handle_execed(vio=%p, userdata=%p{%p}) = 0", vio, userdata, *userdata);
 return 0;
}

static int __handle_passfh(struct roar_vio_calls * vio, void ** userdata, struct roar_message * mes) {
 struct roar_vio_calls svio;
 int sock, fh, clientid;
 void ** clientud;

 sock = -1;

 if ( roar_vio_ctl(vio, ROAR_VIO_CTL_GET_FH, &sock) == -1 )
  sock = -1;

 if (sock == -1)
  return -1;

 fh = roar_socket_recv_fh(sock, NULL, NULL);

 if ( fh != -1 ) {
  roar_vio_open_fh_socket(&svio, fh);

  if (*userdata == NULL || (int)mes->stream != savannahice_stream_get_id(*userdata) ) {
   roar_vio_close(&svio);
   return -1;
  }

  clientid = savannahice_network_add_client_fh(g_network, fh, ROAR_PROTO_ROARAUDIO);
  if ( clientid == -1 ) {
   roar_vio_close(&svio);
   return -1;
  }

  savannahice_network_client_get_userdata(g_network, clientid, &clientud);

  *clientud = *userdata;
  *userdata = NULL;

  return __handle_exec_stream(&svio, clientud, mes);
 } else {
  return -1;
 }
}

static int __handle_kick(void ** userdata, struct roar_message * mes) {
 uint16_t * info = (uint16_t *) mes->data;

 if ( *userdata == NULL )
  return -1;

 if ( mes->datalen != 4 )
  return -1;

 info[0] = ROAR_NET2HOST16(info[0]);
 info[1] = ROAR_NET2HOST16(info[1]);

 if ( info[0] != ROAR_OT_STREAM )
  return -1;

 if ( savannahice_stream_get_id(*userdata) != (int)info[1] )
  return -1;

 savannahice_stream_unref(*userdata);
 *userdata = NULL;

 return 0;
}

static int __handle_caps(struct roar_message * req, struct roar_message * ans) {
 uint16_t flags;

 // check if header is complet.
 if ( req->datalen < 4 )
  return -1;

 // check if version is supported.
 if ( req->data[0] != 0 )
  return -1;

 // check if type is supported.
 if ( req->data[1] != ROAR_CT_STANDARDS )
  return -1;

 flags = (req->data[2] << 8) | req->data[3];

 ans->cmd = ROAR_CMD_OK;

 if ( !(flags & ROAR_CF_REQUEST) ) {
  ans->datalen = 0;
  return 0;
 }

 ans->datalen = 4 + 4;

 ans->data[0] = 0; // version
 ans->data[1] = ROAR_CT_STANDARDS;

 // clear flags:
 ans->data[2] = 0;
 ans->data[3] = 0;

 // RoarAudio Message protocol Version 0:
 ans->data[4] = 0;
 ans->data[5] = 0;
 ans->data[6] = 0;
 ans->data[7] = 0;

 return 0;
}

#ifdef SAVABBAHICE_VERSION
static int __handle_server_info(struct roar_message * req, struct roar_message * ans) {
 const char * version = SAVABBAHICE_VERSION;
 const size_t len = strlen(version);

 // test length:
 if ( req->datalen != 4 )
  return -1;

 // test version:
 if ( req->data[0] != 0 || req->data[1] != 0 )
  return -1;

 // test type (0x0001 = IT_SERVER):
 if ( req->data[2] != 0 || req->data[3] != 1 )
  return -1;

 ans->cmd = ROAR_CMD_OK;

 ans->datalen = 8 + len;
 // version:
 ans->data[0] = 0x00;
 // reserved:
 ans->data[1] = 0x00;
 // indexlen:
 ans->data[2] = 0x00;
 ans->data[3] = 0x01;

 // index 0:
 // - reserved
 ans->data[4] = 0x00;
 // - type (0x00 = ITST_VERSION)
 ans->data[5] = 0x00;
 // - len
 ans->data[6] = (len & 0xFF00) >> 8;
 ans->data[7] = (len & 0x00FF);

 memcpy(&(ans->data[8]), version, len);

 return 0;
}
#endif

static int __handle_gettimeofday(struct roar_message * ans) {
 uint32_t * data = (uint32_t *)ans->data;
 int32_t now = time(NULL);

 ans->datalen = 8*4;
 ans->cmd = ROAR_CMD_OK;

 data[0] = 0;
 data[1] = 0;
 data[2] = now < 0 ? 0xFFFFFFFFLU : 0LU;
 data[3] = htonl((uint32_t)now);
 data[4] = 0;
 data[5] = 0;
 data[6] = 0;
 data[7] = htonl(1000000000UL);

 return 0;
}

static int unset_proto(int client, struct roar_vio_calls * vio, struct roar_buffer ** obuffer, void ** userdata, const struct roar_keyval * para, ssize_t paralen, struct roar_dl_librarypara * pluginpara) {
 (void)client, (void)vio, (void)obuffer, (void)para, (void)paralen, (void)pluginpara;

 if ( *userdata != NULL ) {
  savannahice_stream_unref(*userdata);
  *userdata = NULL;
 }

 return 0;
}

static int handle(int client, struct roar_vio_calls * vio, struct roar_buffer ** obuffer, void ** userdata, const struct roar_keyval * para, ssize_t paralen, struct roar_dl_librarypara * pluginpara) {
 struct roar_message req, ans;
 char * data = NULL;
 int ret = 0;

 (void)obuffer, (void)para, (void)paralen, (void)pluginpara;

 if ( *userdata != NULL )
  if ( savannahice_stream_get_status(*userdata) == ROAR_STREAMSTATE_OLD )
   return __handle_execed(vio, userdata);

 if ( roar_vrecv_message(vio, &req, &data) == -1 )
  return -1;

 memset(&ans, 0, sizeof(ans));

 ans.cmd = ROAR_CMD_ERROR;

 switch (req.cmd) {
  case ROAR_CMD_NOOP:
  case ROAR_CMD_IDENTIFY:
  case ROAR_CMD_AUTH:
  case ROAR_CMD_ATTACH:
    ans.cmd = ROAR_CMD_OK;
   break;
  case ROAR_CMD_QUIT:
    ans.cmd = ROAR_CMD_OK;
    ret = -1; // terminate the connection.
   break;
  case ROAR_CMD_NEW_STREAM:
    if ( __handle_new_stream(client, userdata, &req) == 0 ) {
     ans.cmd = ROAR_CMD_OK;
     ans.stream = savannahice_stream_get_id(*userdata);
    }
   break;
  case ROAR_CMD_EXEC_STREAM:
    if ( __handle_exec_stream(vio, userdata, &req) == 0 )
     ans.cmd = ROAR_CMD_OK;
   break;
  case ROAR_CMD_PASSFH:
    if ( __handle_passfh(vio, userdata, &req) == -1 )
     return -1;
    ans.cmd = ROAR_CMD_OK;
   break;
  case ROAR_CMD_WHOAMI:
    ans.cmd = ROAR_CMD_OK;
    ans.datalen = 1;
    ans.data[0] = client;
   break;
  case ROAR_CMD_KICK:
    if ( __handle_kick(userdata, &req) == 0 )
     ans.cmd = ROAR_CMD_OK;
   break;
  case ROAR_CMD_CAPS:
    __handle_caps(&req, &ans);
   break;
#ifdef SAVABBAHICE_VERSION
  case ROAR_CMD_SERVER_INFO:
    __handle_server_info(&req, &ans);
   break;
#endif
  case ROAR_CMD_GETTIMEOFDAY:
    __handle_gettimeofday(&ans);
   break;
  case ROAR_CMD_EXIT:
    if ( req.datalen == 1 )
     if ( req.data[0] )
      if ( savannahice_network_del_listen_all(g_network) == 0 )
       ans.cmd = ROAR_CMD_OK;
   break;
 }

 roar_vsend_message(vio, &ans, NULL);

 if ( data != NULL )
  roar_mm_free(data);

 return ret;
}

static int status(int client, struct roar_vio_calls * vio, struct roar_buffer ** obuffer, void ** userdata, const struct roar_keyval * para, ssize_t paralen, struct roar_dl_librarypara * pluginpara) {
 ssize_t sizereq;

 (void)client, (void)vio, (void)obuffer, (void)para, (void)paralen, (void)pluginpara;

 if ( *userdata == NULL )
  return ROAR_DL_PROTO_STATUS_RX_READY;

 if ( savannahice_stream_get_status(*userdata) != ROAR_STREAMSTATE_OLD )
  return ROAR_DL_PROTO_STATUS_RX_READY;

 sizereq = savannahice_stream_get_bufferspace(*userdata);

 if ( sizereq != 0 )
  return ROAR_DL_PROTO_STATUS_RX_READY;

 return SAVANNAHICE_CLIENT_STATUS_BUSY;
}

const struct roar_dl_proto g_proto_roar_impl = {
 .proto = ROAR_PROTO_ROARAUDIO,
 .description = "The RoarAudio protocol",
 .set_proto = NULL,
 .unset_proto = unset_proto,
 .handle = handle,
 .flush = NULL,
 .flushed = NULL,
 .status = status
};

//ll
